@@ -19,7 +19,6 @@
 
 #include "rsync.h"
 
-extern int am_server;
 extern int dry_run;
 extern int verbose;
 
@@ -106,14 +105,14 @@ int check_hard_link(struct file_struct *file)
 #if SUPPORT_HARD_LINKS
 static void hard_link_one(int i)
 {
-	struct stat st1,st2;
+	STRUCT_STAT st1,st2;
 
 	if (link_stat(f_name(&hlink_list[i-1]),&st1) != 0) return;
 
 	if (link_stat(f_name(&hlink_list[i]),&st2) != 0) {
 		if (do_link(f_name(&hlink_list[i-1]),f_name(&hlink_list[i])) != 0) {
 			if (verbose > 0)
-				fprintf(FINFO,"link %s => %s : %s\n",
+				rprintf(FINFO,"link %s => %s : %s\n",
 					f_name(&hlink_list[i]),
 					f_name(&hlink_list[i-1]),strerror(errno));
 			return;
@@ -124,14 +123,14 @@ static void hard_link_one(int i)
 		if (do_unlink(f_name(&hlink_list[i])) != 0 ||
 		    do_link(f_name(&hlink_list[i-1]),f_name(&hlink_list[i])) != 0) {
 			if (verbose > 0)
-				fprintf(FINFO,"link %s => %s : %s\n",
+				rprintf(FINFO,"link %s => %s : %s\n",
 					f_name(&hlink_list[i]),
 					f_name(&hlink_list[i-1]),strerror(errno));
 			return;
 		}
 	}
 	if (verbose > 0)
-		fprintf(FINFO,"%s => %s\n",
+		rprintf(FINFO,"%s => %s\n",
 			f_name(&hlink_list[i]),f_name(&hlink_list[i-1]));
 }
 #endif
