@@ -27,18 +27,24 @@
 static int64 total_written;
 static int64 total_read;
 
+static int io_multiplexing_out;
+static int io_multiplexing_in;
+static int multiplex_in_fd;
+static int multiplex_out_fd;
+static time_t last_io;
+
 extern int verbose;
 extern int sparse_files;
 extern int io_timeout;
 
 int64 write_total(void)
 {
-  return total_written;
+	return total_written;
 }
 
 int64 read_total(void)
 {
-  return total_read;
+	return total_read;
 }
 
 static int buffer_f_in = -1;
@@ -49,6 +55,26 @@ void setup_nonblocking(int f_in,int f_out)
 	buffer_f_in = f_in;
 }
 
+static void check_timeout(void)
+{
+	time_t t;
+	
+	if (!io_timeout) return;
+
+	if (!last_io) {
+		last_io = time(NULL);
+		return;
+	}
+
+	t = time(NULL);
+
+	if (last_io && io_timeout && (t-last_io)>io_timeout) {
+		rprintf(FERROR,"read timeout after %d second - exiting\n", 
+			(int)(t-last_io));
+		exit_cleanup(1);
+	}
+}
+
 
 static char *read_buffer;
 static char *read_buffer_p;
@@ -56,6 +82,94 @@ static int read_buffer_len;
 static int read_buffer_size;
 
 
+/* continue trying to read len bytes - don't return until len
+   has been read */
+static void read_loop(int fd, char *buf, int len)
+{
+	while (len) {
+		int n = read(fd, buf, len);
+		if (n > 0) {
+			buf += n;
+			len -= n;
+		}
+		if (n == 0) {
+			rprintf(FERROR,"EOF in read_loop\n");
+			exit_cleanup(1);
+		}
+		if (n == -1) {
+			fd_set fds;
+			struct timeval tv;
+
+			if (errno != EAGAIN && errno != EWOULDBLOCK) {
+				rprintf(FERROR,"io error: %s\n", 
+					strerror(errno));
+				exit_cleanup(1);
+			}
+
+			FD_ZERO(&fds);
+			FD_SET(fd, &fds);
+			tv.tv_sec = io_timeout;
+			tv.tv_usec = 0;
+
+			if (select(fd+1, &fds, NULL, NULL, 
+				   io_timeout?&tv:NULL) != 1) {
+				check_timeout();
+			}
+		}
+	}
+}
+
+static int read_unbuffered(int fd, char *buf, int len)
+{
+	static int remaining;
+	char ibuf[4];
+	int tag, ret=0;
+	char line[1024];
+
+	if (!io_multiplexing_in || fd != multiplex_in_fd) 
+		return read(fd, buf, len);
+
+	while (ret == 0) {
+		if (remaining) {
+			len = MIN(len, remaining);
+			read_loop(fd, buf, len);
+			remaining -= len;
+			ret = len;
+			continue;
+		}
+
+		read_loop(fd, ibuf, 4);
+		tag = IVAL(ibuf, 0);
+
+		remaining = tag & 0xFFFFFF;
+		tag = tag >> 24;
+
+		if (tag == MPLEX_BASE) continue;
+
+		tag -= MPLEX_BASE;
+
+		if (tag != FERROR && tag != FINFO) {
+			rprintf(FERROR,"unexpected tag %d\n", tag);
+			exit_cleanup(1);
+		}
+
+		if (remaining > sizeof(line)-1) {
+			rprintf(FERROR,"multiplexing overflow %d\n\n", 
+				remaining);
+			exit_cleanup(1);
+		}
+
+		read_loop(fd, line, remaining);
+		line[remaining] = 0;
+
+		rprintf(tag,"%s", line);
+		remaining = 0;
+	}
+
+	return ret;
+}
+
+
 /* This function was added to overcome a deadlock problem when using
  * ssh.  It looks like we can't allow our receive queue to get full or
  * ssh will clag up. Uggh.  */
@@ -93,35 +207,12 @@ static void read_check(int f)
 		read_buffer_p = read_buffer;      
 	}
 
-	n = read(f,read_buffer+read_buffer_len,n);
+	n = read_unbuffered(f,read_buffer+read_buffer_len,n);
 	if (n > 0) {
 		read_buffer_len += n;
 	}
 }
 
-static time_t last_io;
-
-
-static void check_timeout(void)
-{
-	time_t t;
-	
-	if (!io_timeout) return;
-
-	if (!last_io) {
-		last_io = time(NULL);
-		return;
-	}
-
-	t = time(NULL);
-
-	if (last_io && io_timeout && (t-last_io)>io_timeout) {
-		rprintf(FERROR,"read timeout after %d second - exiting\n", 
-			(int)(t-last_io));
-		exit_cleanup(1);
-	}
-}
-
 static int readfd(int fd,char *buffer,int N)
 {
 	int  ret;
@@ -141,7 +232,9 @@ static int readfd(int fd,char *buffer,int N)
 			continue;
 		} 
 
-		while ((ret = read(fd,buffer + total,N-total)) == -1) {
+		io_flush();
+
+		while ((ret = read_unbuffered(fd,buffer + total,N-total)) == -1) {
 			fd_set fds;
 
 			if (errno != EAGAIN && errno != EWOULDBLOCK)
@@ -174,7 +267,7 @@ int32 read_int(int f)
   char b[4];
   if ((ret=readfd(f,b,4)) != 4) {
     if (verbose > 1) 
-      rprintf(FERROR,"(%d) Error reading %d bytes : %s\n",
+      rprintf(FERROR,"(%d) read_int: Error reading %d bytes : %s\n",
 	      getpid(),4,ret==-1?strerror(errno):"EOF");
     exit_cleanup(1);
   }
@@ -198,7 +291,7 @@ int64 read_longint(int f)
 	if (remote_version >= 16) {
 		if ((ret=readfd(f,b,8)) != 8) {
 			if (verbose > 1) 
-				rprintf(FERROR,"(%d) Error reading %d bytes : %s\n",
+				rprintf(FERROR,"(%d) read_longint: Error reading %d bytes : %s\n",
 					getpid(),8,ret==-1?strerror(errno):"EOF");
 			exit_cleanup(1);
 		}
@@ -215,7 +308,7 @@ void read_buf(int f,char *buf,int len)
   int ret;
   if ((ret=readfd(f,buf,len)) != len) {
     if (verbose > 1) 
-      rprintf(FERROR,"(%d) Error reading %d bytes : %s\n",
+      rprintf(FERROR,"(%d) read_buf: Error reading %d bytes : %s\n",
 	      getpid(),len,ret==-1?strerror(errno):"EOF");
     exit_cleanup(1);
   }
@@ -306,103 +399,120 @@ int write_file(int f,char *buf,int len)
 
 static int writefd_unbuffered(int fd,char *buf,int len)
 {
-  int total = 0;
-  fd_set w_fds, r_fds;
-  int fd_count, count, got_select=0;
-  struct timeval tv;
+	int total = 0;
+	fd_set w_fds, r_fds;
+	int fd_count, count, got_select=0;
+	struct timeval tv;
 
-  if (buffer_f_in == -1) 
-    return write(fd,buf,len);
+	while (total < len) {
+		int ret = write(fd,buf+total,len-total);
 
-  while (total < len) {
-    int ret = write(fd,buf+total,len-total);
+		if (ret == 0) return total;
 
-    if (ret == 0) return total;
+		if (ret == -1 && !(errno == EWOULDBLOCK || errno == EAGAIN)) 
+			return -1;
 
-    if (ret == -1 && !(errno == EWOULDBLOCK || errno == EAGAIN)) 
-      return -1;
+		if (ret == -1 && got_select) {
+			/* hmmm, we got a write select on the fd and
+			   then failed to write.  Why doesn't that
+			   mean that the fd is dead? It doesn't on
+			   some systems it seems (eg. IRIX) */
+			u_sleep(1000);
+		}
 
-    if (ret == -1 && got_select) {
-	    /* hmmm, we got a write select on the fd and then failed to write.
-	       Why doesn't that mean that the fd is dead? It doesn't on some
-	       systems it seems (eg. IRIX) */
-	    u_sleep(1000);
-#if 0
-	    rprintf(FERROR,"write exception\n");
-	    exit_cleanup(1);
-#endif
-    }
-
-    got_select = 0;
+		got_select = 0;
 
 
-    if (ret == -1) {
-	    if (read_buffer_len < MAX_READ_BUFFER)
-		    read_check(buffer_f_in);
+		if (ret != -1) {
+			total += ret;
+			continue;
+		}
 
-	    fd_count = fd+1;
-	    FD_ZERO(&w_fds);
-	    FD_ZERO(&r_fds);
-	    FD_SET(fd,&w_fds);
-	    if (buffer_f_in != -1) {
-		    FD_SET(buffer_f_in,&r_fds);
-		    if (buffer_f_in > fd) 
-			    fd_count = buffer_f_in+1;
-	    }
+		if (read_buffer_len < MAX_READ_BUFFER && buffer_f_in != -1)
+			read_check(buffer_f_in);
 
-	    tv.tv_sec = BLOCKING_TIMEOUT;
-	    tv.tv_usec = 0;
-	    count = select(fd_count,buffer_f_in == -1? NULL: &r_fds,
-			   &w_fds,NULL,&tv);
+		fd_count = fd+1;
+		FD_ZERO(&w_fds);
+		FD_ZERO(&r_fds);
+		FD_SET(fd,&w_fds);
+		if (buffer_f_in != -1) {
+			FD_SET(buffer_f_in,&r_fds);
+			if (buffer_f_in > fd) 
+				fd_count = buffer_f_in+1;
+		}
 
-	    if (count == -1 && errno != EINTR) {
-		    if (verbose > 1) 
-			    rprintf(FERROR,"select error: %s\n", strerror(errno));
-		    exit_cleanup(1);
-	    }
+		tv.tv_sec = BLOCKING_TIMEOUT;
+		tv.tv_usec = 0;
+		count = select(fd_count,buffer_f_in == -1? NULL: &r_fds,
+			       &w_fds,NULL,&tv);
+		
+		if (count == -1 && errno != EINTR) {
+			if (verbose > 1) 
+				rprintf(FERROR,"select error: %s\n", strerror(errno));
+			exit_cleanup(1);
+		}
+		
+		if (count == 0) {
+			check_timeout();
+			continue;
+		}
+		
+		if (FD_ISSET(fd, &w_fds)) {
+			got_select = 1;
+		}
+	}
 
-	    if (count == 0) {
-		    check_timeout();
-		    continue;
-	    }
-      
-	    if (FD_ISSET(fd, &w_fds)) {
-		    got_select = 1;
-	    }
-    } else {
-	    total += ret;
-    }
-  }
-
-  if (io_timeout)
-	  last_io = time(NULL);
-
-  return total;
+	if (io_timeout)
+		last_io = time(NULL);
+	
+	return total;
 }
 
+
 static char *io_buffer;
 static int io_buffer_count;
 
 void io_start_buffering(int fd)
 {
-	io_buffer = (char *)malloc(IO_BUFFER_SIZE);
+	if (io_buffer) return;
+	multiplex_out_fd = fd;
+	io_buffer = (char *)malloc(IO_BUFFER_SIZE+4);
 	if (!io_buffer) out_of_memory("writefd");
 	io_buffer_count = 0;
+
+	/* leave room for the multiplex header in case it's needed */
+	io_buffer += 4;
+}
+
+void io_flush(void)
+{
+	int fd = multiplex_out_fd;
+	if (!io_buffer_count) return;
+
+	if (io_multiplexing_out) {
+		SIVAL(io_buffer-4, 0, (MPLEX_BASE<<24) + io_buffer_count);
+		if (writefd_unbuffered(fd, io_buffer-4, io_buffer_count+4) !=
+		    io_buffer_count+4) {
+			rprintf(FERROR,"write failed\n");
+			exit_cleanup(1);
+		}
+	} else {
+		if (writefd_unbuffered(fd, io_buffer, io_buffer_count) != 
+		    io_buffer_count) {
+			rprintf(FERROR,"write failed\n");
+			exit_cleanup(1);
+		}
+	}
+	io_buffer_count = 0;
 }
 
 void io_end_buffering(int fd)
 {
-	if (io_buffer_count) {
-		if (writefd_unbuffered(fd, io_buffer, 
-				       io_buffer_count) != 
-		    io_buffer_count) {
-			rprintf(FERROR,"write failed\n");
-			exit_cleanup(1);
-		}
-		io_buffer_count = 0;
+	io_flush();
+	if (!io_multiplexing_out) {
+		free(io_buffer-4);
+		io_buffer = NULL;
 	}
-	free(io_buffer);
-	io_buffer = NULL;
 }
 
 static int writefd(int fd,char *buf,int len1)
@@ -420,14 +530,7 @@ static int writefd(int fd,char *buf,int len1)
 			io_buffer_count += n;
 		}
 		
-		if (io_buffer_count == IO_BUFFER_SIZE) {
-			if (writefd_unbuffered(fd, io_buffer, 
-					       io_buffer_count) != 
-			    io_buffer_count) {
-				return -1;
-			}
-			io_buffer_count = 0;
-		}
+		if (io_buffer_count == IO_BUFFER_SIZE) io_flush();
 	}
 
 	return len1;
@@ -436,15 +539,15 @@ static int writefd(int fd,char *buf,int len1)
 
 void write_int(int f,int32 x)
 {
-  int ret;
-  char b[4];
-  SIVAL(b,0,x);
-  if ((ret=writefd(f,b,4)) != 4) {
-    rprintf(FERROR,"write_int failed : %s\n",
-	    ret==-1?strerror(errno):"EOF");
-    exit_cleanup(1);
-  }
-  total_written += 4;
+	int ret;
+	char b[4];
+	SIVAL(b,0,x);
+	if ((ret=writefd(f,b,4)) != 4) {
+		rprintf(FERROR,"write_int failed : %s\n",
+			ret==-1?strerror(errno):"EOF");
+		exit_cleanup(1);
+	}
+	total_written += 4;
 }
 
 void write_longint(int f, int64 x)
@@ -472,13 +575,13 @@ void write_longint(int f, int64 x)
 
 void write_buf(int f,char *buf,int len)
 {
-  int ret;
-  if ((ret=writefd(f,buf,len)) != len) {
-    rprintf(FERROR,"write_buf failed : %s\n",
-	    ret==-1?strerror(errno):"EOF");
-    exit_cleanup(1);
-  }
-  total_written += len;
+	int ret;
+	if ((ret=writefd(f,buf,len)) != len) {
+		rprintf(FERROR,"write_buf failed : %s\n",
+			ret==-1?strerror(errno):"EOF");
+		exit_cleanup(1);
+	}
+	total_written += len;
 }
 
 /* write a string to the connection */
@@ -533,3 +636,45 @@ void io_printf(int fd, const char *format, ...)
 
 	write_sbuf(fd, buf);
 }
+
+
+/* setup for multiplexing an error stream with the data stream */
+void io_start_multiplex_out(int fd)
+{
+	multiplex_out_fd = fd;
+	io_flush();
+	io_start_buffering(fd);
+	io_multiplexing_out = 1;
+}
+
+/* setup for multiplexing an error stream with the data stream */
+void io_start_multiplex_in(int fd)
+{
+	multiplex_in_fd = fd;
+	io_flush();
+	if (read_buffer_len) {
+		fprintf(stderr,"ERROR: data in read buffer at mplx start\n");
+		exit_cleanup(1);
+	}
+
+	io_multiplexing_in = 1;
+}
+
+/* write an message to the error stream */
+int io_multiplex_write(int f, char *buf, int len)
+{
+	if (!io_multiplexing_out) return 0;
+
+	io_flush();
+
+	SIVAL(io_buffer-4, 0, ((MPLEX_BASE + f)<<24) + len);
+	memcpy(io_buffer, buf, len);
+
+	writefd_unbuffered(multiplex_out_fd, io_buffer-4, len+4);
+	return 1;
+}
+
+void io_close_input(int fd)
+{
+	buffer_f_in = -1;
+}
